from __future__ import annotations

import asyncio
import contextvars
import functools
import io
from collections.abc import Mapping
from typing import Any

from flask import Flask

from .base import HttpClient, RequestData, Response, UploadedFile


class FlaskHttpClient(HttpClient):
    def __init__(self, app: Flask) -> None:
        self.app = app

    def _build_request_kwargs(
        self,
        data: RequestData | None,
        json_data: object | None,
        files: dict[str, UploadedFile] | None,
        **kwargs: Any,
    ) -> dict[str, Any]:
        request_kwargs = dict(kwargs)

        if json_data is not None:
            request_kwargs["json"] = json_data
            return request_kwargs

        if files is None:
            if data is not None:
                request_kwargs["data"] = data
            return request_kwargs

        if data is not None and not isinstance(data, Mapping):
            raise TypeError("Flask multipart requests require mapping form data")

        multipart_data = dict(data or {})
        for key, (filename, content, content_type) in files.items():
            multipart_data[key] = (
                io.BytesIO(content),
                filename,
                content_type or "application/octet-stream",
            )

        request_kwargs["data"] = multipart_data
        return request_kwargs

    def _do_request(
        self,
        url: str,
        method: str,
        headers: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> Response:
        cookies = kwargs.pop("cookies", None)
        files = kwargs.pop("files", None)
        data = kwargs.pop("data", None)
        json_data = kwargs.pop("json", None)
        request_kwargs = self._build_request_kwargs(
            data=data,
            json_data=json_data,
            files=files,
            **kwargs,
        )

        with self.app.test_client() as client:
            for key, value in (cookies or {}).items():
                client.set_cookie(key, value)

            response = client.open(
                url,
                method=method.upper(),
                headers=headers,
                **request_kwargs,
            )

        return Response(
            status_code=response.status_code,
            data=response.data,
            headers=dict(response.headers),
        )

    async def request(
        self,
        url: str,
        method: str,
        headers: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> Response:
        loop = asyncio.get_running_loop()
        ctx = contextvars.copy_context()
        func_call = functools.partial(
            ctx.run,
            self._do_request,
            url=url,
            method=method,
            headers=headers,
            **kwargs,
        )
        return await loop.run_in_executor(None, func_call)


class AsyncFlaskHttpClient(FlaskHttpClient):
    pass



from __future__ import annotations

from collections.abc import Awaitable, Callable, Iterable, Mapping
from json import dumps
from typing import Any
from urllib.parse import urlsplit

from django.core.exceptions import BadRequest, SuspiciousOperation
from django.core.files.uploadedfile import SimpleUploadedFile
from django.http import Http404, HttpRequest, HttpResponse, StreamingHttpResponse
from django.test.client import AsyncRequestFactory, RequestFactory
from django.urls import Resolver404, ResolverMatch, resolve

from .base import HttpClient, RequestData, Response, UploadedFile, merge_cookies


def resolve_view_match(view: object, url: str) -> ResolverMatch | None:
    urlpatterns = getattr(view, "_cross_web_urlpatterns", None)
    if urlpatterns is None:
        return None

    try:
        return resolve(urlsplit(url).path, urlconf=tuple(urlpatterns))
    except Resolver404:
        return None


class DjangoHttpClient(HttpClient):
    def __init__(self, view: Callable[..., HttpResponse]) -> None:
        self.view = view

    def _to_django_headers(self, headers: Mapping[str, str]) -> dict[str, str]:
        return {
            f"HTTP_{key.upper().replace('-', '_')}": value
            for key, value in headers.items()
        }

    def _build_request_data(
        self,
        data: RequestData | None,
        json_data: object | None,
        files: dict[str, UploadedFile] | None,
        headers: dict[str, str] | None,
    ) -> tuple[object | None, dict[str, Any]]:
        request_kwargs: dict[str, Any] = {}
        request_data: object | None = data

        if json_data is not None:
            request_data = dumps(json_data)
            request_kwargs["content_type"] = "application/json"

        if files:
            if request_data is not None and not isinstance(request_data, Mapping):
                raise TypeError("Django multipart requests require mapping form data")

            multipart_data = dict(request_data or {})
            for key, (filename, content, content_type) in files.items():
                multipart_data[key] = SimpleUploadedFile(
                    filename,
                    content,
                    content_type=content_type or "application/octet-stream",
                )
            request_data = multipart_data
        elif (
            request_data is not None
            and isinstance(request_data, bytes)
            and headers is not None
            and "Content-Type" in headers
        ):
            request_kwargs["content_type"] = headers["Content-Type"]

        return request_data, request_kwargs

    async def _do_request(self, request: HttpRequest) -> Response:
        try:
            resolver_match = request.resolver_match
            if resolver_match is None:
                response = self.view(request)
            else:
                response = self.view(
                    request,
                    *resolver_match.args,
                    **resolver_match.kwargs,
                )
        except Http404:
            return Response(status_code=404, data=b"Not found")
        except (BadRequest, SuspiciousOperation) as exc:
            return Response(status_code=400, data=str(exc).encode())

        return Response(
            status_code=response.status_code,
            data=response.content,
            headers=dict(response.headers),
        )

    async def request(
        self,
        url: str,
        method: str,
        headers: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> Response:
        cookies = kwargs.pop("cookies", None)
        files = kwargs.pop("files", None)
        data = kwargs.pop("data", None)
        json_data = kwargs.pop("json", None)

        merged_headers = merge_cookies(headers, cookies) or {}
        request_data, request_kwargs = self._build_request_data(
            data=data,
            json_data=json_data,
            files=files,
            headers=merged_headers,
        )
        request_kwargs.update(self._to_django_headers(merged_headers))
        request_kwargs.update(kwargs)

        request_factory = RequestFactory()
        request = getattr(request_factory, method)(
            url, data=request_data, **request_kwargs
        )
        request.resolver_match = resolve_view_match(self.view, url)
        if cookies is not None:
            request.COOKIES = dict(cookies)

        return await self._do_request(request)


class AsyncDjangoHttpClient(HttpClient):
    def __init__(self, view: Callable[..., Awaitable[HttpResponse]]) -> None:
        self.view = view

    def _to_django_headers(self, headers: Mapping[str, str]) -> dict[str, str]:
        return {
            f"HTTP_{key.upper().replace('-', '_')}": value
            for key, value in headers.items()
        }

    def _build_request_data(
        self,
        data: RequestData | None,
        json_data: object | None,
        files: dict[str, UploadedFile] | None,
        headers: dict[str, str] | None,
    ) -> tuple[object | None, dict[str, Any]]:
        request_kwargs: dict[str, Any] = {}
        request_data: object | None = data

        if json_data is not None:
            request_data = dumps(json_data)
            request_kwargs["content_type"] = "application/json"

        if files:
            if request_data is not None and not isinstance(request_data, Mapping):
                raise TypeError("Django multipart requests require mapping form data")

            multipart_data = dict(request_data or {})
            for key, (filename, content, content_type) in files.items():
                multipart_data[key] = SimpleUploadedFile(
                    filename,
                    content,
                    content_type=content_type or "application/octet-stream",
                )
            request_data = multipart_data
        elif (
            request_data is not None
            and isinstance(request_data, bytes)
            and headers is not None
            and "Content-Type" in headers
        ):
            request_kwargs["content_type"] = headers["Content-Type"]

        return request_data, request_kwargs

    async def _do_request(self, request: HttpRequest) -> Response:
        try:
            resolver_match = request.resolver_match
            if resolver_match is None:
                response = await self.view(request)
            else:
                response = await self.view(
                    request,
                    *resolver_match.args,
                    **resolver_match.kwargs,
                )
        except Http404:
            return Response(status_code=404, data=b"Not found")
        except (BadRequest, SuspiciousOperation) as exc:
            return Response(status_code=400, data=str(exc).encode())

        if isinstance(response, StreamingHttpResponse):
            content = response.streaming_content
            data = b""
            if isinstance(content, Iterable):
                data = b"".join(content)
        else:
            data = response.content

        return Response(
            status_code=response.status_code,
            data=data,
            headers=dict(response.headers),
        )

    async def request(
        self,
        url: str,
        method: str,
        headers: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> Response:
        cookies = kwargs.pop("cookies", None)
        files = kwargs.pop("files", None)
        data = kwargs.pop("data", None)
        json_data = kwargs.pop("json", None)

        merged_headers = merge_cookies(headers, cookies) or {}
        request_data, request_kwargs = self._build_request_data(
            data=data,
            json_data=json_data,
            files=files,
            headers=merged_headers,
        )
        request_kwargs.update(self._to_django_headers(merged_headers))
        request_kwargs.update(kwargs)

        request_factory = AsyncRequestFactory()
        request = getattr(request_factory, method)(
            url, data=request_data, **request_kwargs
        )
        request.resolver_match = resolve_view_match(self.view, url)
        if cookies is not None:
            request.COOKIES = dict(cookies)

        return await self._do_request(request)



from __future__ import annotations

from typing import TYPE_CHECKING, Mapping, Optional, cast

from ._base import AsyncHTTPRequestAdapter, FormData, HTTPMethod, QueryParams

if TYPE_CHECKING:
    from starlette.requests import Request


class StarletteRequestAdapter(AsyncHTTPRequestAdapter):
    def __init__(self, request: Request) -> None:
        self._request = request
        # Starlette Headers are case-insensitive Mapping
        self._headers: Optional[Mapping[str, str]] = None

    @property
    def method(self) -> HTTPMethod:
        # Starlette method is already uppercase string
        return cast(HTTPMethod, self._request.method)

    @property
    def query_params(self) -> QueryParams:
        # Starlette QueryParams behaves like a MultiDict Mapping
        return cast(QueryParams, self._request.query_params)

    @property
    def path_params(self) -> Mapping[str, str]:
        return cast(Mapping[str, str], self._request.path_params)

    @property
    def headers(self) -> Mapping[str, str]:
        # Cache the immutable headers object for direct access
        # Starlette Headers are case-insensitive
        if self._headers is None:
            self._headers = self._request.headers

        return self._headers

    @property
    def content_type(self) -> Optional[str]:
        # Access directly via headers property for consistency
        return self.headers.get("content-type")

    async def get_body(self) -> bytes:
        return await self._request.body()

    async def get_form_data(self) -> FormData:
        multipart_data = await self._request.form()

        return FormData(
            files=multipart_data,
            form=multipart_data,
        )

    @property
    def url(self) -> str:
        return str(self._request.url)

    @property
    def cookies(self) -> Mapping[str, str]:
        return cast(Mapping[str, str], self._request.cookies)



from __future__ import annotations

from io import BytesIO
from typing import TYPE_CHECKING, Any, Mapping, Optional, cast

from ._base import AsyncHTTPRequestAdapter, FormData, HTTPMethod, QueryParams

if TYPE_CHECKING:
    from aiohttp import web


class AiohttpHTTPRequestAdapter(AsyncHTTPRequestAdapter):
    def __init__(
        self,
        request: web.Request,
        body: Optional[bytes] = None,
        form_data: Optional[FormData] = None,
    ) -> None:
        self.request = request
        self._body = body
        self._form_data = form_data

    @classmethod
    async def create(cls, request: web.Request) -> "AiohttpHTTPRequestAdapter":
        """Create an adapter and pre-read the body to avoid PayloadAccessError"""
        content_type = request.headers.get("content-type", "")
        form_data = None
        body = None

        if content_type.startswith("multipart/form-data"):
            # Pre-process multipart data
            reader = await request.multipart()
            data: dict[str, Any] = {}
            files: dict[str, Any] = {}

            while field := await reader.next():
                from aiohttp.multipart import BodyPartReader

                assert isinstance(field, BodyPartReader)
                assert field.name

                if field.filename:
                    files[field.name] = BytesIO(await field.read(decode=False))
                else:
                    data[field.name] = await field.text()

            form_data = FormData(files=files, form=data)
        else:
            # For non-multipart requests, read the body
            body = await request.read()

        return cls(request, body, form_data)

    @property
    def query_params(self) -> QueryParams:
        return cast(QueryParams, self.request.query.copy())  # type: ignore[attr-defined]

    @property
    def path_params(self) -> Mapping[str, Any]:
        return cast(Mapping[str, Any], dict(self.request.match_info))

    async def get_body(self) -> bytes:
        if self._form_data is not None:
            return b""
        if self._body is None:
            self._body = await self.request.read()
        return self._body

    @property
    def method(self) -> HTTPMethod:
        return cast("HTTPMethod", self.request.method.upper())

    @property
    def headers(self) -> Mapping[str, str]:
        return self.request.headers

    async def get_form_data(self) -> FormData:
        if self._form_data is not None:
            return self._form_data

        if self.content_type and self.content_type.startswith("multipart/form-data"):
            # Process multipart data
            reader = await self.request.multipart()
            data: dict[str, Any] = {}
            files: dict[str, Any] = {}

            while field := await reader.next():
                from aiohttp.multipart import BodyPartReader

                assert isinstance(field, BodyPartReader)
                assert field.name

                if field.filename:
                    files[field.name] = BytesIO(await field.read(decode=False))
                else:
                    data[field.name] = await field.text()

            self._form_data = FormData(files=files, form=data)
            return self._form_data
        else:
            # For URL-encoded form data
            post_data = await self.request.post()
            return FormData(files={}, form=dict(post_data))

    @property
    def content_type(self) -> Optional[str]:
        return self.headers.get("content-type")

    @property
    def url(self) -> str:
        return str(self.request.url)

    @property
    def cookies(self) -> Mapping[str, str]:
        return self.request.cookies



from __future__ import annotations

import json
from dataclasses import dataclass
from typing import TYPE_CHECKING, Literal, Mapping, List, Union, cast
from typing_extensions import Self
from urllib.parse import urlencode

if TYPE_CHECKING:
    from fastapi import Response as FastAPIResponse

JsonType = Union[
    str, int, float, bool, None, Mapping[str, "JsonType"], List["JsonType"]
]


@dataclass
class Cookie:
    name: str
    value: str
    secure: bool
    path: str | None = None
    domain: str | None = None
    max_age: int | None = None
    httponly: bool = True
    samesite: Literal["lax", "strict", "none"] = "lax"


@dataclass
class Response:
    status_code: int
    body: str | None = None
    cookies: list[Cookie] | None = None
    headers: Mapping[str, str] | None = None

    @classmethod
    def redirect(
        cls,
        url: str,
        query_params: Mapping[str, str | list[str]] | None = None,
        headers: Mapping[str, str] | None = None,
        cookies: list[Cookie] | None = None,
    ) -> Self:
        headers = headers or {}

        if query_params:
            url = url + "?" + urlencode(query_params)

        return cls(
            status_code=302,
            headers={"Location": url, **headers},
            cookies=cookies,
        )

    def json(self) -> JsonType:
        if self.body is None:
            return None

        return cast(JsonType, json.loads(self.body))

    def to_fastapi(self) -> FastAPIResponse:
        from fastapi import Response as FastAPIResponse

        response = FastAPIResponse(
            status_code=self.status_code,
            headers=self.headers,
            content=self.body,
        )

        for cookie in self.cookies or []:
            response.set_cookie(
                cookie.name,
                cookie.value,
                secure=cookie.secure,
                path=cookie.path,
                domain=cookie.domain,
                max_age=cookie.max_age,
                httponly=cookie.httponly,
                samesite=cookie.samesite,
            )

        return response



from __future__ import annotations

import io
from collections.abc import Mapping
from typing import Any

from quart import Quart
from quart.datastructures import FileStorage

from .base import HttpClient, RequestData, Response, UploadedFile


class QuartHttpClient(HttpClient):
    def __init__(self, app: Quart) -> None:
        self.app = app

    def _build_request_kwargs(
        self,
        data: RequestData | None,
        json_data: object | None,
        files: dict[str, UploadedFile] | None,
        **kwargs: Any,
    ) -> dict[str, Any]:
        request_kwargs = dict(kwargs)

        if json_data is not None:
            request_kwargs["json"] = json_data
            return request_kwargs

        if files is not None:
            if data is not None and not isinstance(data, Mapping):
                raise TypeError("Quart multipart requests require mapping form data")

            request_kwargs["form"] = dict(data or {})
            request_kwargs["files"] = {
                key: FileStorage(
                    stream=io.BytesIO(content),
                    filename=filename,
                    content_type=content_type or "application/octet-stream",
                )
                for key, (filename, content, content_type) in files.items()
            }
            return request_kwargs

        if isinstance(data, Mapping):
            request_kwargs["form"] = dict(data)
        elif data is not None:
            request_kwargs["data"] = data

        return request_kwargs

    async def request(
        self,
        url: str,
        method: str,
        headers: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> Response:
        cookies = kwargs.pop("cookies", None)
        files = kwargs.pop("files", None)
        data = kwargs.pop("data", None)
        json_data = kwargs.pop("json", None)
        request_kwargs = self._build_request_kwargs(
            data=data,
            json_data=json_data,
            files=files,
            **kwargs,
        )

        async with self.app.test_app() as test_app, self.app.app_context():
            client = test_app.test_client()
            for key, value in (cookies or {}).items():
                client.set_cookie("localhost", key, value)

            response = await getattr(client, method)(
                url,
                headers=headers,
                **request_kwargs,
            )

        return Response(
            status_code=response.status_code,
            data=await response.get_data(),
            headers=dict(response.headers),
        )



from __future__ import annotations

from typing import Any

from litestar import Litestar
from litestar.testing import TestClient

from .base import HttpClient, Response


class LitestarHttpClient(HttpClient):
    def __init__(self, app: Litestar) -> None:
        self.app = app

    async def request(
        self,
        url: str,
        method: str,
        headers: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> Response:
        cookies = kwargs.pop("cookies", None)

        with TestClient(app=self.app) as client:
            client.cookies.update(dict(cookies or {}))
            response = client.request(
                method.upper(),
                url,
                headers=headers,
                **kwargs,
            )

        return Response(
            status_code=response.status_code,
            data=response.content,
            headers=dict(response.headers),
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Mapping, Optional, cast

from ._base import AsyncHTTPRequestAdapter, FormData, HTTPMethod, QueryParams

if TYPE_CHECKING:
    from litestar import Request


class LitestarRequestAdapter(AsyncHTTPRequestAdapter):
    def __init__(self, request: Request[Any, Any, Any]) -> None:
        self.request = request

    @property
    def query_params(self) -> QueryParams:
        return self.request.query_params

    @property
    def path_params(self) -> Mapping[str, Any]:
        return cast(Mapping[str, Any], self.request.path_params)

    @property
    def method(self) -> HTTPMethod:
        return cast("HTTPMethod", self.request.method.upper())

    @property
    def headers(self) -> Mapping[str, str]:
        return self.request.headers

    @property
    def content_type(self) -> Optional[str]:
        content_type, params = self.request.content_type

        # combine content type and params
        if params:
            content_type += "; " + "; ".join(f"{k}={v}" for k, v in params.items())

        return content_type

    async def get_body(self) -> bytes:
        return await self.request.body()

    async def get_form_data(self) -> FormData:
        multipart_data = await self.request.form()

        return FormData(form=multipart_data, files=multipart_data)

    @property
    def url(self) -> str:
        return str(self.request.url)

    @property
    def cookies(self) -> Mapping[str, str]:
        return self.request.cookies



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Mapping, Optional, Union, cast

from ._base import AsyncHTTPRequestAdapter, FormData, HTTPMethod, QueryParams

if TYPE_CHECKING:
    from sanic.request import File, Request


def convert_request_to_files_dict(request: Request) -> dict[str, Any]:
    """Converts the request.files dictionary to a dictionary of sanic Request objects.

    `request.files` has the following format, even if only a single file is uploaded:

    ```python
    {
        "textFile": [
            sanic.request.File(type="text/plain", body=b"strawberry", name="textFile.txt")
        ]
    }
    ```

    Note that the dictionary entries are lists.
    """
    request_files = cast("Optional[dict[str, list[File]]]", request.files)

    if not request_files:
        return {}

    files_dict: dict[str, Union[File, list[File]]] = {}

    for field_name, file_list in request_files.items():
        assert len(file_list) == 1

        files_dict[field_name] = file_list[0]

    return files_dict


class SanicHTTPRequestAdapter(AsyncHTTPRequestAdapter):
    def __init__(self, request: Request) -> None:
        self.request = request

    @property
    def query_params(self) -> QueryParams:
        # Just a heads up, Sanic's request.args uses urllib.parse.parse_qs
        # to parse query string parameters. This returns a dictionary where
        # the keys are the unique variable names and the values are lists
        # of values for each variable name. To ensure consistency, we're
        # enforcing the use of the first value in each list.
        args = self.request.get_args(keep_blank_values=True)
        return {k: args.get(k, None) for k in args}

    @property
    def path_params(self) -> Mapping[str, Any]:
        return cast(Mapping[str, Any], self.request.match_info or {})

    @property
    def method(self) -> HTTPMethod:
        return cast("HTTPMethod", self.request.method.upper())

    @property
    def headers(self) -> Mapping[str, str]:
        return self.request.headers

    @property
    def content_type(self) -> Optional[str]:
        return self.request.content_type

    async def get_body(self) -> bytes:
        return self.request.body

    async def get_form_data(self) -> FormData:
        assert self.request.form is not None

        files = convert_request_to_files_dict(self.request)

        return FormData(form=self.request.form, files=files)

    @property
    def url(self) -> str:
        return self.request.url

    @property
    def cookies(self) -> Mapping[str, str]:
        return self.request.cookies



from collections.abc import Mapping
from typing import Optional, Protocol, Union

from .request._base import HTTPMethod, PathParams


class BaseRequestProtocol(Protocol):
    """Protocol defining the minimal interface for HTTP requests."""

    @property
    def query_params(self) -> Mapping[str, Optional[Union[str, list[str]]]]: ...

    @property
    def path_params(self) -> PathParams: ...

    @property
    def method(self) -> HTTPMethod: ...

    @property
    def headers(self) -> Mapping[str, str]: ...


__all__ = ["BaseRequestProtocol"]



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Mapping, Optional, Union, cast

from ._base import (
    AsyncHTTPRequestAdapter,
    FormData,
    HTTPMethod,
    QueryParams,
    SyncHTTPRequestAdapter,
)

if TYPE_CHECKING:
    from django.http import HttpRequest


class DjangoHTTPRequestAdapter(SyncHTTPRequestAdapter):
    def __init__(self, request: HttpRequest) -> None:
        self.request = request

    @property
    def query_params(self) -> QueryParams:
        return cast(QueryParams, self.request.GET.dict())

    @property
    def path_params(self) -> Mapping[str, Any]:
        if self.request.resolver_match is None:
            return {}
        return cast(Mapping[str, Any], self.request.resolver_match.kwargs)

    @property
    def body(self) -> Union[str, bytes]:
        return cast(Union[str, bytes], self.request.body.decode())

    @property
    def method(self) -> HTTPMethod:
        assert self.request.method is not None

        return cast("HTTPMethod", self.request.method.upper())

    @property
    def headers(self) -> Mapping[str, str]:
        return cast(Mapping[str, str], self.request.headers)

    @property
    def post_data(self) -> Mapping[str, Union[str, bytes]]:
        return cast(Mapping[str, Union[str, bytes]], self.request.POST)

    @property
    def files(self) -> Mapping[str, Any]:
        return cast(Mapping[str, Any], self.request.FILES)

    def get_form_data(self) -> FormData:
        return FormData(
            files=cast(Mapping[str, Any], self.request.FILES),
            form=cast(Mapping[str, Union[str, bytes]], self.request.POST),
        )

    @property
    def content_type(self) -> Optional[str]:
        return cast(Optional[str], self.request.content_type)

    @property
    def url(self) -> str:
        return cast(str, self.request.build_absolute_uri())

    @property
    def cookies(self) -> Mapping[str, str]:
        return cast(Mapping[str, str], self.request.COOKIES)


class AsyncDjangoHTTPRequestAdapter(AsyncHTTPRequestAdapter):
    def __init__(self, request: HttpRequest) -> None:
        self.request = request

    @property
    def query_params(self) -> QueryParams:
        return cast(QueryParams, self.request.GET.dict())

    @property
    def path_params(self) -> Mapping[str, Any]:
        if self.request.resolver_match is None:
            return {}
        return cast(Mapping[str, Any], self.request.resolver_match.kwargs)

    @property
    def method(self) -> HTTPMethod:
        assert self.request.method is not None

        return cast("HTTPMethod", self.request.method.upper())

    @property
    def headers(self) -> Mapping[str, str]:
        return cast(Mapping[str, str], self.request.headers)

    @property
    def content_type(self) -> Optional[str]:
        return self.headers.get("Content-type")

    async def get_body(self) -> bytes:
        return cast(bytes, self.request.body)

    async def get_form_data(self) -> FormData:
        return FormData(
            files=cast(Mapping[str, Any], self.request.FILES),
            form=cast(Mapping[str, Union[str, bytes]], self.request.POST),
        )

    @property
    def url(self) -> str:
        return cast(str, self.request.build_absolute_uri())

    @property
    def cookies(self) -> Mapping[str, str]:
        return cast(Mapping[str, str], self.request.COOKIES)



from __future__ import annotations

import abc
import json
from collections.abc import Mapping
from dataclasses import dataclass
from functools import cached_property
from typing import Any, Literal, Optional, Union, cast

JSON = Union[dict[str, "JSON"], list["JSON"], str, int, float, bool, None]
RequestData = Union[bytes, str, Mapping[str, object]]
UploadedFile = tuple[str, bytes, Optional[str]]
RequestMethod = Literal["head", "get", "post", "patch", "put", "delete"]


@dataclass
class Response:
    status_code: int
    data: bytes

    def __init__(
        self, status_code: int, data: bytes, *, headers: dict[str, str] | None = None
    ) -> None:
        self.status_code = status_code
        self.data = data
        self._headers = headers or {}

    @cached_property
    def headers(self) -> Mapping[str, str]:
        return {key.lower(): value for key, value in self._headers.items()}

    @property
    def text(self) -> str:
        return self.data.decode()

    @property
    def json(self) -> JSON:
        return cast(JSON, json.loads(self.data))


class HttpClient(abc.ABC):
    supports_form_data = True

    @abc.abstractmethod
    def __init__(self, app: Any) -> None:
        raise NotImplementedError

    @abc.abstractmethod
    async def request(
        self,
        url: str,
        method: RequestMethod,
        headers: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> Response:
        raise NotImplementedError

    async def get(
        self,
        url: str,
        headers: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> Response:
        return await self.request(url, "get", headers=headers, **kwargs)

    async def post(
        self,
        url: str,
        data: RequestData | None = None,
        json: JSON | None = None,
        files: dict[str, UploadedFile] | None = None,
        headers: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> Response:
        return await self.request(
            url,
            "post",
            headers=headers,
            data=data,
            json=json,
            files=files,
            **kwargs,
        )


def merge_cookies(
    headers: dict[str, str] | None, cookies: Mapping[str, str] | None
) -> dict[str, str] | None:
    if not cookies:
        return headers

    merged_headers = dict(headers or {})
    existing_cookie_header = merged_headers.pop("Cookie", "")
    existing_cookies: dict[str, str] = {}

    for cookie in existing_cookie_header.split(";"):
        name, separator, value = cookie.strip().partition("=")
        if separator and name:
            existing_cookies[name] = value

    combined_cookies = {**existing_cookies, **cookies}
    merged_headers["Cookie"] = "; ".join(
        f"{name}={value}" for name, value in combined_cookies.items()
    )
    return merged_headers



from __future__ import annotations

from collections.abc import Mapping
from typing import Any

from aiohttp import FormData, web
from aiohttp.test_utils import TestClient, TestServer

from .base import HttpClient, RequestData, Response, UploadedFile


class AiohttpHttpClient(HttpClient):
    def __init__(self, app: web.Application) -> None:
        self.app = app

    def _build_data(
        self,
        data: RequestData | None,
        files: dict[str, UploadedFile] | None,
    ) -> RequestData | FormData | None:
        if not files:
            return data

        if data is not None and not isinstance(data, Mapping):
            raise TypeError("aiohttp multipart requests require mapping form data")

        form_data = FormData()

        for key, value in (data or {}).items():
            form_data.add_field(key, str(value))

        for key, (filename, content, content_type) in files.items():
            form_data.add_field(
                key,
                content,
                filename=filename,
                content_type=content_type or "application/octet-stream",
            )

        return form_data

    async def request(
        self,
        url: str,
        method: str,
        headers: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> Response:
        cookies = kwargs.pop("cookies", None)
        files = kwargs.pop("files", None)
        data = kwargs.pop("data", None)

        if files is not None or data is not None:
            kwargs["data"] = self._build_data(data, files)

        if cookies is not None:
            kwargs["cookies"] = dict(cookies)

        async with TestClient(TestServer(self.app)) as client:
            response = await getattr(client, method)(url, headers=headers, **kwargs)

            return Response(
                status_code=response.status,
                data=await response.read(),
                headers=dict(response.headers),
            )



from .exceptions import HTTPException
from .protocols import BaseRequestProtocol
from .request import AsyncHTTPRequest
from .request._aiohttp import AiohttpHTTPRequestAdapter
from .request._base import AsyncHTTPRequestAdapter, FormData, SyncHTTPRequestAdapter
from .request._chalice import ChaliceHTTPRequestAdapter
from .request._django import AsyncDjangoHTTPRequestAdapter, DjangoHTTPRequestAdapter
from .request._flask import AsyncFlaskHTTPRequestAdapter, FlaskHTTPRequestAdapter
from .request._litestar import LitestarRequestAdapter
from .request._quart import QuartHTTPRequestAdapter
from .request._sanic import SanicHTTPRequestAdapter
from .request._starlette import StarletteRequestAdapter
from .request._testing import TestingRequestAdapter
from .response import Cookie, Response

__all__ = [
    "AiohttpHTTPRequestAdapter",
    "AsyncDjangoHTTPRequestAdapter",
    "AsyncFlaskHTTPRequestAdapter",
    "AsyncHTTPRequest",
    "AsyncHTTPRequestAdapter",
    "BaseRequestProtocol",
    "ChaliceHTTPRequestAdapter",
    "Cookie",
    "DjangoHTTPRequestAdapter",
    "FlaskHTTPRequestAdapter",
    "FormData",
    "HTTPException",
    "LitestarRequestAdapter",
    "QuartHTTPRequestAdapter",
    "Response",
    "SanicHTTPRequestAdapter",
    "StarletteRequestAdapter",
    "SyncHTTPRequestAdapter",
    "TestingRequestAdapter",
]



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Mapping, Optional, Union, cast

from ._base import FormData, HTTPMethod, QueryParams, SyncHTTPRequestAdapter

if TYPE_CHECKING:
    from chalice.app import Request


class ChaliceHTTPRequestAdapter(SyncHTTPRequestAdapter):
    def __init__(self, request: Request) -> None:
        self.request = request

    @property
    def query_params(self) -> QueryParams:
        return self.request.query_params or {}

    @property
    def path_params(self) -> Mapping[str, Any]:
        return cast(Mapping[str, Any], self.request.uri_params or {})

    @property
    def body(self) -> Union[str, bytes]:
        return self.request.raw_body

    @property
    def method(self) -> HTTPMethod:
        return cast("HTTPMethod", self.request.method.upper())

    @property
    def headers(self) -> Mapping[str, str]:
        return self.request.headers

    @property
    def post_data(self) -> Mapping[str, Union[str, bytes]]:
        # Chalice doesn't support traditional form data
        raise NotImplementedError("Chalice does not support form data")

    @property
    def files(self) -> Mapping[str, Any]:
        # Chalice doesn't support file uploads out of the box
        raise NotImplementedError("Chalice does not support file uploads")

    def get_form_data(self) -> FormData:
        # Chalice doesn't support form data
        raise NotImplementedError("Chalice does not support form data")

    @property
    def content_type(self) -> Optional[str]:
        return self.request.headers.get("Content-Type", None)

    @property
    def url(self) -> str:
        # Construct URL from context
        context = self.request.context
        stage = context.get("stage", "")
        domain = context.get("domainName", "")
        path = context.get("path", "")

        # Build the URL
        protocol = "https"  # API Gateway typically uses HTTPS
        if stage and stage != "prod":
            url = f"{protocol}://{domain}/{stage}{path}"
        else:
            url = f"{protocol}://{domain}{path}"

        # Add query string if present
        if self.request.query_params:
            from urllib.parse import urlencode

            query_string = urlencode(self.request.query_params)
            url = f"{url}?{query_string}"

        return url

    @property
    def cookies(self) -> Mapping[str, str]:
        # Chalice doesn't have direct cookie support
        # Cookies would come in the Cookie header
        cookie_header = self.request.headers.get("Cookie", "")
        if not cookie_header:
            return {}

        cookies = {}
        for cookie in cookie_header.split(";"):
            cookie = cookie.strip()
            if "=" in cookie:
                name, value = cookie.split("=", 1)
                cookies[name] = value

        return cookies



from __future__ import annotations

from typing import Any

from .base import HttpClient, Response, merge_cookies


class SanicHttpClient(HttpClient):
    def __init__(self, app: Any) -> None:
        self.app = app

    async def request(
        self,
        url: str,
        method: str,
        headers: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> Response:
        cookies = kwargs.pop("cookies", None)
        request_headers = merge_cookies(headers, cookies)

        _, response = await self.app.asgi_client.request(
            method.upper(),
            url,
            headers=request_headers,
            **kwargs,
        )

        return Response(
            status_code=response.status,
            data=response.body,
            headers=dict(response.headers),
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Mapping, Optional, Any

from typing_extensions import Self

if TYPE_CHECKING:
    from starlette.requests import Request as StarletteRequest

from ._base import (
    AsyncHTTPRequestAdapter,
    FormData,
    HTTPMethod,
    PathParams,
    QueryParams,
)
from ._starlette import StarletteRequestAdapter
from ._testing import TestingRequestAdapter


class AsyncHTTPRequest:
    def __init__(self, adapter: AsyncHTTPRequestAdapter) -> None:
        self._adapter = adapter

    @classmethod
    def from_starlette(cls, request: StarletteRequest) -> Self:
        adapter = StarletteRequestAdapter(request)

        return cls(adapter)

    @classmethod
    def from_fastapi(cls, request: StarletteRequest) -> Self:
        return cls.from_starlette(request)

    @classmethod
    def from_django(cls, request: Any) -> Self:
        # Import here to avoid circular imports and optional Django dependency
        from ._django import AsyncDjangoHTTPRequestAdapter

        adapter = AsyncDjangoHTTPRequestAdapter(request)
        return cls(adapter)

    @classmethod
    def from_flask(cls, request: Any) -> Self:
        # Import here to avoid circular imports and optional Flask dependency
        from ._flask import AsyncFlaskHTTPRequestAdapter

        adapter = AsyncFlaskHTTPRequestAdapter(request)
        return cls(adapter)

    @classmethod
    def from_sanic(cls, request: Any) -> Self:
        # Import here to avoid circular imports and optional Sanic dependency
        from ._sanic import SanicHTTPRequestAdapter

        adapter = SanicHTTPRequestAdapter(request)
        return cls(adapter)

    @classmethod
    def from_aiohttp(cls, request: Any) -> Self:
        # Import here to avoid circular imports and optional aiohttp dependency
        from ._aiohttp import AiohttpHTTPRequestAdapter

        adapter = AiohttpHTTPRequestAdapter(request)
        return cls(adapter)

    @classmethod
    def from_quart(cls, request: Any) -> Self:
        # Import here to avoid circular imports and optional Quart dependency
        from ._quart import QuartHTTPRequestAdapter

        adapter = QuartHTTPRequestAdapter(request)
        return cls(adapter)

    @classmethod
    def from_litestar(cls, request: Any) -> Self:
        # Import here to avoid circular imports and optional Litestar dependency
        from ._litestar import LitestarRequestAdapter

        adapter = LitestarRequestAdapter(request)
        return cls(adapter)

    @classmethod
    def from_form_data(cls, data: Mapping[str, str]) -> Self:
        adapter = TestingRequestAdapter(
            form_data=FormData(files={}, form=data),
            content_type="application/x-www-form-urlencoded",
        )
        return cls(adapter)

    @property
    def method(self) -> HTTPMethod:
        """The HTTP method of the request."""
        return self._adapter.method

    @property
    def query_params(self) -> QueryParams:
        """The query parameters of the request."""
        return self._adapter.query_params

    @property
    def path_params(self) -> PathParams:
        """The path parameters of the request."""
        return self._adapter.path_params

    @property
    def headers(self) -> Mapping[str, str]:
        """The request headers (case-insensitive keys recommended)."""
        return self._adapter.headers

    @property
    def content_type(self) -> Optional[str]:
        """The 'Content-Type' header value, if present."""
        return self._adapter.content_type

    @property
    def url(self) -> str:
        """The URL of the request."""
        return self._adapter.url

    @property
    def cookies(self) -> Mapping[str, str]:
        """The request cookies."""
        return self._adapter.cookies

    async def get_body(self) -> bytes:
        """Return the raw request body as bytes."""
        return await self._adapter.get_body()

    async def get_form_data(self) -> FormData:
        """
        Return parsed form data (multipart/form-data or application/x-www-form-urlencoded).
        """
        return await self._adapter.get_form_data()



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Mapping, Optional, Union, cast

from ._base import (
    AsyncHTTPRequestAdapter,
    FormData,
    HTTPMethod,
    QueryParams,
    SyncHTTPRequestAdapter,
)

if TYPE_CHECKING:
    from flask import Request


class FlaskHTTPRequestAdapter(SyncHTTPRequestAdapter):
    def __init__(self, request: Request) -> None:
        self.request = request

    @property
    def query_params(self) -> QueryParams:
        return self.request.args.to_dict()

    @property
    def path_params(self) -> Mapping[str, Any]:
        return cast(Mapping[str, Any], self.request.view_args or {})

    @property
    def body(self) -> Union[str, bytes]:
        return self.request.data.decode()

    @property
    def method(self) -> HTTPMethod:
        return cast("HTTPMethod", self.request.method.upper())

    @property
    def headers(self) -> Mapping[str, str]:
        return self.request.headers  # type: ignore

    @property
    def post_data(self) -> Mapping[str, Union[str, bytes]]:
        return self.request.form

    @property
    def files(self) -> Mapping[str, Any]:
        return self.request.files

    def get_form_data(self) -> FormData:
        return FormData(
            files=self.request.files,
            form=self.request.form,
        )

    @property
    def content_type(self) -> Optional[str]:
        return self.request.content_type

    @property
    def url(self) -> str:
        return self.request.url

    @property
    def cookies(self) -> Mapping[str, str]:
        return self.request.cookies


class AsyncFlaskHTTPRequestAdapter(AsyncHTTPRequestAdapter):
    def __init__(self, request: Request) -> None:
        self.request = request

    @property
    def query_params(self) -> QueryParams:
        return self.request.args.to_dict()

    @property
    def path_params(self) -> Mapping[str, Any]:
        return cast(Mapping[str, Any], self.request.view_args or {})

    @property
    def method(self) -> HTTPMethod:
        return cast("HTTPMethod", self.request.method.upper())

    @property
    def content_type(self) -> Optional[str]:
        return self.request.content_type

    @property
    def headers(self) -> Mapping[str, str]:
        return self.request.headers  # type: ignore

    async def get_body(self) -> bytes:
        return self.request.data

    async def get_form_data(self) -> FormData:
        return FormData(
            files=self.request.files,
            form=self.request.form,
        )

    @property
    def url(self) -> str:
        return self.request.url

    @property
    def cookies(self) -> Mapping[str, str]:
        return self.request.cookies



from __future__ import annotations

from typing import TYPE_CHECKING, Mapping, Optional, cast

from ._base import AsyncHTTPRequestAdapter, FormData, HTTPMethod, QueryParams

if TYPE_CHECKING:
    from quart import Request


class QuartHTTPRequestAdapter(AsyncHTTPRequestAdapter):
    def __init__(self, request: Request) -> None:
        self.request = request

    @property
    def query_params(self) -> QueryParams:
        return self.request.args.to_dict()

    @property
    def path_params(self) -> Mapping[str, str]:
        return cast(Mapping[str, str], self.request.view_args or {})

    @property
    def method(self) -> HTTPMethod:
        return cast("HTTPMethod", self.request.method.upper())

    @property
    def content_type(self) -> Optional[str]:
        return self.request.content_type

    @property
    def headers(self) -> Mapping[str, str]:
        return self.request.headers  # type: ignore

    async def get_body(self) -> bytes:
        return await self.request.data

    async def get_form_data(self) -> FormData:
        files = await self.request.files
        form = await self.request.form
        return FormData(files=files, form=form)

    @property
    def url(self) -> str:
        return self.request.url

    @property
    def cookies(self) -> Mapping[str, str]:
        return self.request.cookies



from __future__ import annotations

import json
from typing import Any

from chalice.app import Chalice
from chalice.test import Client

from .base import HttpClient, Response, merge_cookies


class ChaliceHttpClient(HttpClient):
    supports_form_data = False

    def __init__(self, app: Chalice) -> None:
        self.app = app

    async def request(
        self,
        url: str,
        method: str,
        headers: dict[str, str] | None = None,
        **kwargs: Any,
    ) -> Response:
        cookies = kwargs.pop("cookies", None)
        files = kwargs.pop("files", None)
        data = kwargs.pop("data", None)
        json_data = kwargs.pop("json", None)

        if files is not None:
            raise NotImplementedError("Chalice test client does not support files")

        request_headers = merge_cookies(headers, cookies)
        body: str | bytes | None = None

        if json_data is not None:
            body = json.dumps(json_data)
            request_headers = dict(request_headers or {})
            request_headers.setdefault("Content-Type", "application/json")
        elif isinstance(data, (bytes, str)):
            body = data
        elif data is not None:
            raise NotImplementedError("Chalice test client does not support form data")

        request_kwargs = {"headers": request_headers, **kwargs}
        if body is not None:
            request_kwargs["body"] = body

        with Client(self.app) as client:
            response = getattr(client.http, method)(url, **request_kwargs)

        response_body = response.body
        if isinstance(response_body, str):
            response_body = response_body.encode()

        return Response(
            status_code=response.status_code,
            data=response_body,
            headers=dict(response.headers),
        )



from __future__ import annotations

import json
from typing import Any, Mapping, Optional

from ._base import AsyncHTTPRequestAdapter, FormData, HTTPMethod, QueryParams


class TestingRequestAdapter(AsyncHTTPRequestAdapter):
    __test__ = False

    def __init__(
        self,
        *,
        method: HTTPMethod = "POST",
        query_params: QueryParams | None = None,
        path_params: Mapping[str, Any] | None = None,
        headers: Mapping[str, str] | None = None,
        content_type: str | None = None,
        url: str = "http://testserver/",
        cookies: Mapping[str, str] | None = None,
        form_data: FormData | None = None,
        json: dict[str, Any] | None = None,
    ) -> None:
        self._method = method
        self._query_params = query_params or {}
        self._path_params = path_params or {}
        self._headers = headers or {}
        self._content_type = content_type
        self._url = url
        self._cookies = cookies or {}
        self._form_data = form_data or FormData(files={}, form={})
        self._json = json

    @property
    def method(self) -> HTTPMethod:
        return self._method

    @property
    def query_params(self) -> QueryParams:
        return self._query_params

    @property
    def path_params(self) -> Mapping[str, Any]:
        return self._path_params

    @property
    def headers(self) -> Mapping[str, str]:
        return self._headers

    @property
    def content_type(self) -> Optional[str]:
        return self._content_type

    async def get_body(self) -> bytes:
        if self._json is not None:
            return json.dumps(self._json).encode("utf-8")

        return b""

    async def get_form_data(self) -> FormData:
        return self._form_data

    @property
    def url(self) -> str:
        return self._url

    @property
    def cookies(self) -> Mapping[str, str]:
        return self._cookies



import abc
from collections.abc import Mapping
from dataclasses import dataclass
from typing import Any, Literal, Optional, Union

HTTPMethod = Literal[
    "GET", "POST", "PUT", "DELETE", "PATCH", "HEAD", "OPTIONS", "TRACE"
]

QueryParams = Mapping[str, Optional[str]]
PathParams = Mapping[str, Any]


@dataclass
class FormData:
    files: Mapping[str, Any]
    form: Mapping[str, Any]

    def get(self, key: str) -> Any:
        return self.form.get(key)


class SyncHTTPRequestAdapter(abc.ABC):
    """
    Abstract Base Class defining the interface for accessing HTTP request data
    in a framework-agnostic way for synchronous operations.
    """

    @property
    @abc.abstractmethod
    def method(self) -> HTTPMethod:
        """The HTTP method of the request (e.g., 'GET', 'POST')."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def query_params(self) -> QueryParams:
        """The query parameters of the request."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def path_params(self) -> PathParams:
        """The path parameters of the request."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def headers(self) -> Mapping[str, str]:
        """The request headers. Header names should ideally be case-insensitive."""
        # Note: Real implementations might need to handle case-insensitivity
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def content_type(self) -> Optional[str]:
        """The 'Content-Type' header value, if present."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def body(self) -> Union[str, bytes]:
        """Return the raw request body as bytes or string."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def post_data(self) -> Mapping[str, Union[str, bytes]]:
        """Return the parsed POST data as a mapping of field names to values."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def files(self) -> Mapping[str, Any]:
        """Return uploaded files from the request."""
        raise NotImplementedError

    @abc.abstractmethod
    def get_form_data(self) -> FormData:
        """
        Return parsed form data (multipart/form-data or application/x-www-form-urlencoded).
        Returns an empty FormData object if the content type is not form data
        or if parsing fails.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def url(self) -> str:
        """The URL of the request."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def cookies(self) -> Mapping[str, str]:
        """The request cookies."""
        raise NotImplementedError


class AsyncHTTPRequestAdapter(abc.ABC):
    """
    Abstract Base Class defining the interface for accessing HTTP request data
    in a framework-agnostic way.
    """

    @property
    @abc.abstractmethod
    def method(self) -> HTTPMethod:
        """The HTTP method of the request (e.g., 'GET', 'POST')."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def query_params(self) -> QueryParams:
        """The query parameters of the request."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def path_params(self) -> PathParams:
        """The path parameters of the request."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def headers(self) -> Mapping[str, str]:
        """The request headers. Header names should ideally be case-insensitive."""
        # Note: Real implementations might need to handle case-insensitivity
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def content_type(self) -> Optional[str]:
        """The 'Content-Type' header value, if present."""
        raise NotImplementedError

    @abc.abstractmethod
    async def get_body(self) -> bytes:
        """Return the raw request body as bytes."""
        raise NotImplementedError

    @abc.abstractmethod
    async def get_form_data(self) -> FormData:
        """
        Return parsed form data (multipart/form-data or application/x-www-form-urlencoded).
        Returns an empty FormData object if the content type is not form data
        or if parsing fails.
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def url(self) -> str:
        """The URL of the request."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def cookies(self) -> Mapping[str, str]:
        """The request cookies."""
        raise NotImplementedError
